@@ -65,16 +65,14 @@ typedef boost::container::vector<ProcessPtr> ProcessList;
 
 /**
  * Represents an application process, as spawned by a SpawningKit::Spawner. Every
- * Process has a PID, an admin socket and a list of sockets on which it listens for
- * connections. A Process object is contained inside a Group.
+ * Process has a PID, a stdin pipe, an output pipe and a list of sockets on which
+ * it listens for connections. A Process object is contained inside a Group.
  *
- * The admin socket, an anonymous Unix domain socket, is mapped to the process's
- * STDIN and STDOUT and has two functions.
+ * The stdin pipe is mapped to the process's STDIN and is used for garbage
+ * collection: closing the STDIN part causes the process to gracefully terminate itself.
  *
- *  1. It acts as the main communication channel with the process. Commands are
- *     sent to and responses are received from it.
- *  2. It's used for garbage collection: closing the STDIN part causes the process
- *     to gracefully terminate itself.
+ * The output pipe is mapped to the process' STDOUT and STDERR. All data coming
+ * from those pipes will be printed.
  *
  * Except for the otherwise documented parts, this class is not thread-safe,
  * so only use within the Pool lock.
@@ -102,7 +100,7 @@ typedef boost::container::vector<ProcessPtr> ProcessList;
  */
 class Process {
 public:
-	static const unsigned int MAX_SESSION_SOCKETS = 3;
+	static const unsigned int MAX_SOCKETS_ACCEPTING_HTTP_REQUESTS = 3;
 
 private:
 	/*************************************************************
@@ -121,21 +119,20 @@ private:
 	int concurrency;
 
 	/**
-	 * A subset of 'sockets': all sockets that speak the
-	 * "session" or "http_session" protocol.
+	 * A subset of 'sockets': all sockets that accept HTTP requests
+	 * from the Passenger Core controller.
 	 */
-	unsigned int sessionSocketCount;
-	Socket *sessionSockets[MAX_SESSION_SOCKETS];
+	unsigned int socketsAcceptingHttpRequestsCount;
+	Socket *socketsAcceptingHttpRequests[MAX_SOCKETS_ACCEPTING_HTTP_REQUESTS];
 
-	/** Admin socket. See Process class description. */
-	FileDescriptor adminSocket;
+	/** Input pipe. See Process class description. */
+	FileDescriptor inputPipe;
 
 	/**
-	 * Pipe on which this process outputs errors. Mapped to the process's STDERR.
-	 * Only Processes spawned by DirectSpawner have this set.
-	 * SmartSpawner-spawned Processes use the same STDERR as their parent preloader processes.
+	 * Pipe on which this process outputs stdout and stderr data. Mapped to the
+	 * process's STDOUT and STDERR.
 	 */
-	FileDescriptor errorPipe;
+	FileDescriptor outputPipe;
 
 	/**
 	 * The code revision of the application, inferred through various means.
@@ -201,9 +198,9 @@ private:
 		};
 
 		struct SocketStringOffsets {
-			String name;
 			String address;
 			String protocol;
+			String description;
 		};
 
 		vector<SocketStringOffsets> socketStringOffsets;
@@ -211,15 +208,33 @@ private:
 	};
 
 	void appendJsonFieldToBuffer(std::string &buffer, const Json::Value &json,
-		const char *key, InitializationLog::String &str) const
+		const char *key, InitializationLog::String &str, bool required = true) const
 	{
-		StaticString value = getJsonStaticStringField(json, key);
+		StaticString value;
+		if (required) {
+			value = getJsonStaticStringField(json, key);
+		} else {
+			value = getJsonStaticStringField(json, Json::StaticString(key),
+				StaticString());
+		}
 		str.offset = buffer.size();
 		str.size   = value.size();
 		buffer.append(value.data(), value.size());
 		buffer.append(1, '\0');
 	}
 
+	void initializeSocketsAndStringFields(const SpawningKit::Result &result) {
+		Json::Value doc, sockets(Json::arrayValue);
+		vector<SpawningKit::Result::Socket>::const_iterator it, end = result.sockets.end();
+
+		for (it = result.sockets.begin(); it != end; it++) {
+			sockets.append(it->inspectAsJson());
+		}
+
+		doc["sockets"] = sockets;
+		initializeSocketsAndStringFields(doc);
+	}
+
 	void initializeSocketsAndStringFields(const Json::Value &json) {
 		InitializationLog log;
 		string buffer;
@@ -238,9 +253,10 @@ private:
 			const Json::Value &socket = *it;
 			InitializationLog::SocketStringOffsets offsets;
 
-			appendJsonFieldToBuffer(buffer, socket, "name", offsets.name);
 			appendJsonFieldToBuffer(buffer, socket, "address", offsets.address);
 			appendJsonFieldToBuffer(buffer, socket, "protocol", offsets.protocol);
+			appendJsonFieldToBuffer(buffer, socket, "description", offsets.description,
+				false);
 
 			log.socketStringOffsets.push_back(offsets);
 		}
@@ -267,13 +283,14 @@ private:
 			const Json::Value &socket = *it;
 			this->sockets.add(
 				info.pid,
-				StaticString(base + log.socketStringOffsets[i].name.offset,
-					log.socketStringOffsets[i].name.size),
 				StaticString(base + log.socketStringOffsets[i].address.offset,
 					log.socketStringOffsets[i].address.size),
 				StaticString(base + log.socketStringOffsets[i].protocol.offset,
 					log.socketStringOffsets[i].protocol.size),
-				getJsonIntField(socket, "concurrency")
+				StaticString(base + log.socketStringOffsets[i].description.offset,
+					log.socketStringOffsets[i].description.size),
+				getJsonIntField(socket, "concurrency"),
+				getJsonBoolField(socket, "accept_http_requests")
 			);
 		}
 
@@ -283,44 +300,51 @@ private:
 		}
 	}
 
-	void indexSessionSockets() {
+	void indexSocketsAcceptingHttpRequests() {
 		SocketList::iterator it;
 
 		concurrency = 0;
-		memset(sessionSockets, 0, sizeof(sessionSockets));
+		memset(socketsAcceptingHttpRequests, 0, sizeof(socketsAcceptingHttpRequests));
 
 		for (it = sockets.begin(); it != sockets.end(); it++) {
 			Socket *socket = &(*it);
-			if (socket->protocol == "session" || socket->protocol == "http_session") {
-				if (sessionSocketCount == MAX_SESSION_SOCKETS) {
-					throw RuntimeException("The process has too many session sockets. "
-						"A maximum of " + toString(MAX_SESSION_SOCKETS) + " is allowed");
-				}
-				sessionSockets[sessionSocketCount] = socket;
-				sessionSocketCount++;
+			if (!socket->acceptHttpRequests) {
+				continue;
+			}
+			if (socketsAcceptingHttpRequestsCount == MAX_SOCKETS_ACCEPTING_HTTP_REQUESTS) {
+				throw RuntimeException("The process has too many sockets that accept HTTP requests. "
+					"A maximum of " + toString(MAX_SOCKETS_ACCEPTING_HTTP_REQUESTS) + " is allowed");
+			}
 
-				if (concurrency != -1) {
-					if (socket->concurrency == 0) {
-						// If one of the sockets has a concurrency of
-						// 0 (unlimited) then we mark this entire Process
-						// as having a concurrency of 0.
-						concurrency = -1;
-					} else {
-						concurrency += socket->concurrency;
-					}
+			socketsAcceptingHttpRequests[socketsAcceptingHttpRequestsCount] = socket;
+			socketsAcceptingHttpRequestsCount++;
+
+			if (concurrency >= 0) {
+				if (socket->concurrency < 0) {
+					// If one of the sockets has a concurrency of
+					// < 0 (unknown) then we mark this entire Process
+					// as having a concurrency of -1 (unknown).
+					concurrency = -1;
+				} else if (socket->concurrency == 0) {
+					// If one of the sockets has a concurrency of
+					// 0 (unlimited) then we mark this entire Process
+					// as having a concurrency of 0.
+					concurrency = -999;
+				} else {
+					concurrency += socket->concurrency;
 				}
 			}
 		}
 
-		if (concurrency == -1) {
+		if (concurrency == -999) {
 			concurrency = 0;
 		}
 	}
 
 	void destroySelf() const {
 		this->~Process();
-		LockGuard l(getContext()->getMmSyncher());
-		getContext()->getProcessObjectPool().free(const_cast<Process *>(this));
+		LockGuard l(getContext()->memoryManagementSyncher);
+		getContext()->processObjectPool.free(const_cast<Process *>(this));
 	}
 
 
@@ -353,7 +377,8 @@ private:
 		return result;
 	}
 
-	SpawningKit::PipeWatcherPtr makePipeWatcher(const SpawningKit::ConfigPtr &config, FileDescriptor socket, const char *channel, pid_t pid, const BasicGroupInfo *groupInfo);
+	string getAppGroupName(const BasicGroupInfo *info) const;
+	string getAppLogFile(const BasicGroupInfo *info) const;
 
 public:
 	/*************************************************************
@@ -427,13 +452,13 @@ public:
 	ProcessMetrics metrics;
 
 
-	Process(const BasicGroupInfo *groupInfo, const Json::Value &json)
-		: info(this, groupInfo, json),
-		  sessionSocketCount(0),
-		  spawnerCreationTime(getJsonUint64Field(json, "spawner_creation_time")),
-		  spawnStartTime(getJsonUint64Field(json, "spawn_start_time")),
+	Process(const BasicGroupInfo *groupInfo, const Json::Value &args)
+		: info(this, groupInfo, args),
+		  socketsAcceptingHttpRequestsCount(0),
+		  spawnerCreationTime(getJsonUint64Field(args, "spawner_creation_time")),
+		  spawnStartTime(getJsonUint64Field(args, "spawn_start_time")),
 		  spawnEndTime(SystemTime::getUsec()),
-		  dummy(json["type"] == "dummy"),
+		  dummy(args["type"] == "dummy"),
 		  requiresShutdown(false),
 		  refcount(1),
 		  index(-1),
@@ -447,21 +472,46 @@ public:
 		  longRunningConnectionsAborted(false),
 		  shutdownStartTime(0)
 	{
-		initializeSocketsAndStringFields(json);
-		indexSessionSockets();
+		initializeSocketsAndStringFields(args);
+		indexSocketsAcceptingHttpRequests();
+	}
 
-		const SpawningKit::Result *skResult = dynamic_cast<const SpawningKit::Result *>(&json);
-		if (skResult != NULL) {
-			adminSocket = skResult->adminSocket;
-			errorPipe = skResult->errorPipe;
+	Process(const BasicGroupInfo *groupInfo, const SpawningKit::Result &skResult,
+		const Json::Value &args)
+		: info(this, groupInfo, skResult),
+		  socketsAcceptingHttpRequestsCount(0),
+		  spawnerCreationTime(getJsonUint64Field(args, "spawner_creation_time")),
+		  spawnStartTime(skResult.spawnStartTime),
+		  spawnEndTime(skResult.spawnEndTime),
+		  dummy(skResult.dummy),
+		  requiresShutdown(false),
+		  refcount(1),
+		  index(-1),
+		  lastUsed(spawnEndTime),
+		  sessions(0),
+		  processed(0),
+		  lifeStatus(ALIVE),
+		  enabled(ENABLED),
+		  oobwStatus(OOBW_NOT_ACTIVE),
+		  m_osProcessExists(true),
+		  longRunningConnectionsAborted(false),
+		  shutdownStartTime(0)
+	{
+		initializeSocketsAndStringFields(skResult);
+		indexSocketsAcceptingHttpRequests();
 
-			if (adminSocket != -1) {
-				SpawningKit::PipeWatcherPtr watcher = makePipeWatcher(getContext()->getSpawningKitConfig(), adminSocket, "stdout", info.pid, groupInfo);
-			}
+		inputPipe = skResult.stdinFd;
+		outputPipe = skResult.stdoutAndErrFd;
 
-			if (errorPipe != -1) {
-				SpawningKit::PipeWatcherPtr watcher = makePipeWatcher(getContext()->getSpawningKitConfig(), errorPipe, "stderr", info.pid, groupInfo);
+		if (outputPipe != -1) {
+			SpawningKit::PipeWatcherPtr watcher = boost::make_shared<SpawningKit::PipeWatcher>(
+				outputPipe, "output", getAppGroupName(groupInfo),
+				getAppLogFile(groupInfo), skResult.pid);
+			if (!args["log_file"].isNull()) {
+				watcher->setLogFile(args["log_file"].asString());
 			}
+			watcher->initialize();
+			watcher->start();
 		}
 	}
 
@@ -479,8 +529,8 @@ public:
 	void forceMaxConcurrency(int value) {
 		assert(value >= 0);
 		concurrency = value;
-		for (unsigned i = 0; i < sessionSocketCount; i++) {
-			sessionSockets[i]->concurrency = concurrency;
+		for (unsigned i = 0; i < socketsAcceptingHttpRequestsCount; i++) {
+			socketsAcceptingHttpRequests[i]->concurrency = concurrency;
 		}
 	}
 
@@ -553,8 +603,8 @@ public:
 			lifeStatus = SHUTDOWN_TRIGGERED;
 			shutdownStartTime = now;
 		}
-		if (!dummy) {
-			syscalls::shutdown(adminSocket, SHUT_WR);
+		if (inputPipe != -1) {
+			inputPipe.close();
 		}
 	}
 
@@ -644,23 +694,23 @@ public:
 		return sockets;
 	}
 
-	Socket *findSessionSocketWithLowestBusyness() const {
-		if (OXT_UNLIKELY(sessionSocketCount == 0)) {
+	Socket *findSocketsAcceptingHttpRequestsAndWithLowestBusyness() const {
+		if (OXT_UNLIKELY(socketsAcceptingHttpRequestsCount == 0)) {
 			return NULL;
-		} else if (sessionSocketCount == 1) {
-			return sessionSockets[0];
+		} else if (socketsAcceptingHttpRequestsCount == 1) {
+			return socketsAcceptingHttpRequests[0];
 		} else {
-			int leastBusySessionSocketIndex = 0;
-			int lowestBusyness = sessionSockets[0]->busyness();
+			int leastBusySocketIndex = 0;
+			int lowestBusyness = socketsAcceptingHttpRequests[0]->busyness();
 
-			for (unsigned i = 1; i < sessionSocketCount; i++) {
-				if (sessionSockets[i]->busyness() < lowestBusyness) {
-					leastBusySessionSocketIndex = i;
-					lowestBusyness = sessionSockets[i]->busyness();
+			for (unsigned i = 1; i < socketsAcceptingHttpRequestsCount; i++) {
+				if (socketsAcceptingHttpRequests[i]->busyness() < lowestBusyness) {
+					leastBusySocketIndex = i;
+					lowestBusyness = socketsAcceptingHttpRequests[i]->busyness();
 				}
 			}
 
-			return sessionSockets[leastBusySessionSocketIndex];
+			return socketsAcceptingHttpRequests[leastBusySocketIndex];
 		}
 	}
 
@@ -700,15 +750,15 @@ public:
 		/* Different processes within a Group may have different
 		 * 'concurrency' values. We want:
 		 * - the process with the smallest busyness to be be picked for routing.
-		 * - to give processes with concurrency == 0 more priority (in general)
+		 * - to give processes with concurrency == 0 or -1 more priority (in general)
 		 *   over processes with concurrency > 0.
 		 * Therefore, in case of processes with concurrency > 0, we describe our
 		 * busyness as a percentage of 'concurrency', with the percentage value
 		 * in [0..INT_MAX] instead of [0..1]. That way, the busyness value
 		 * of processes with concurrency > 0 is usually higher than that of processes
-		 * with concurrency == 0.
+		 * with concurrency == 0 or -1.
 		 */
-		if (concurrency == 0) {
+		if (concurrency <= 0) {
 			return sessions;
 		} else {
 			return (int) (((long long) sessions * INT_MAX) / (double) concurrency);
@@ -720,7 +770,7 @@ public:
 	 * process.
 	 */
 	bool isTotallyBusy() const {
-		return concurrency != 0 && sessions >= concurrency;
+		return concurrency > 0 && sessions >= concurrency;
 	}
 
 	/**
@@ -745,7 +795,7 @@ public:
 	 * not result in any harmful behavior.
 	 */
 	SessionPtr newSession(unsigned long long now = 0) {
-		Socket *socket = findSessionSocketWithLowestBusyness();
+		Socket *socket = findSocketsAcceptingHttpRequestsAndWithLowestBusyness();
 		if (socket->isTotallyBusy()) {
 			return SessionPtr();
 		} else {
@@ -772,7 +822,7 @@ public:
 
 			~Guard() {
 				if (session != NULL) {
-					context->getSessionObjectPool().free(session);
+					context->sessionObjectPool.free(session);
 				}
 			}
 
@@ -782,8 +832,8 @@ public:
 		};
 
 		Context *context = getContext();
-		LockGuard l(context->getMmSyncher());
-		Session *session = context->getSessionObjectPool().malloc();
+		LockGuard l(context->memoryManagementSyncher);
+		Session *session = context->sessionObjectPool.malloc();
 		Guard guard(context, session);
 		session = new (session) Session(context, &info, socket);
 		guard.clear();
@@ -882,10 +932,13 @@ public:
 			for (it = sockets.begin(); it != sockets.end(); it++) {
 				const Socket &socket = *it;
 				stream << "<socket>";
-				stream << "<name>" << escapeForXml(socket.name) << "</name>";
 				stream << "<address>" << escapeForXml(socket.address) << "</address>";
 				stream << "<protocol>" << escapeForXml(socket.protocol) << "</protocol>";
+				if (!socket.description.empty()) {
+					stream << "<description>" << escapeForXml(socket.description) << "</description>";
+				}
 				stream << "<concurrency>" << socket.concurrency << "</concurrency>";
+				stream << "<accept_http_requests>" << socket.acceptHttpRequests << "</accept_http_requests>";
 				stream << "<sessions>" << socket.sessions << "</sessions>";
 				stream << "</socket>";
 			}
