@@ -1,6 +1,6 @@
 /*
  *  Phusion Passenger - https://www.phusionpassenger.com/
- *  Copyright (c) 2011-2016 Phusion Holding B.V.
+ *  Copyright (c) 2011-2017 Phusion Holding B.V.
  *
  *  "Passenger", "Phusion Passenger" and "Union Station" are registered
  *  trademarks of Phusion Holding B.V.
@@ -393,8 +393,8 @@ Controller::determineHeaderSizeForSessionProtocol(Request *req,
 			}
 		}
 	} else {
-		state.serverName = defaultServerName;
-		state.serverPort = defaultServerPort;
+		state.serverName = req->config->defaultServerName;
+		state.serverPort = req->config->defaultServerPort;
 	}
 
 	dataSize += sizeof("SERVER_NAME");
@@ -404,7 +404,7 @@ Controller::determineHeaderSizeForSessionProtocol(Request *req,
 	dataSize += state.serverPort.size() + 1;
 
 	dataSize += sizeof("SERVER_SOFTWARE");
-	dataSize += serverSoftware.size() + 1;
+	dataSize += req->config->serverSoftware.size() + 1;
 
 	dataSize += sizeof("SERVER_PROTOCOL");
 	dataSize += sizeof("HTTP/1.1");
@@ -446,14 +446,6 @@ Controller::determineHeaderSizeForSessionProtocol(Request *req,
 		dataSize += sizeof("on");
 	}
 
-	if (req->options.analytics) {
-		dataSize += sizeof("PASSENGER_TXN_ID");
-		dataSize += req->options.transaction->getTxnId().size() + 1;
-
-		dataSize += sizeof("PASSENGER_DELTA_MONOTONIC");
-		dataSize += delta_monotonic.size() + 1;
-	}
-
 	if (req->upgraded()) {
 		dataSize += sizeof("HTTP_CONNECTION");
 		dataSize += sizeof("upgrade");
@@ -515,7 +507,7 @@ Controller::constructHeaderForSessionProtocol(Request *req, char * restrict buff
 	pos = appendData(pos, end, "", 1);
 
 	pos = appendData(pos, end, P_STATIC_STRING_WITH_NULL("SERVER_SOFTWARE"));
-	pos = appendData(pos, end, serverSoftware);
+	pos = appendData(pos, end, req->config->serverSoftware);
 	pos = appendData(pos, end, "", 1);
 
 	pos = appendData(pos, end, P_STATIC_STRING_WITH_NULL("SERVER_PROTOCOL"));
@@ -564,16 +556,6 @@ Controller::constructHeaderForSessionProtocol(Request *req, char * restrict buff
 		pos = appendData(pos, end, P_STATIC_STRING_WITH_NULL("on"));
 	}
 
-	if (req->options.analytics) {
-		pos = appendData(pos, end, P_STATIC_STRING_WITH_NULL("PASSENGER_TXN_ID"));
-		pos = appendData(pos, end, req->options.transaction->getTxnId());
-		pos = appendData(pos, end, "", 1);
-
-		pos = appendData(pos, end, P_STATIC_STRING_WITH_NULL("PASSENGER_DELTA_MONOTONIC"));
-		pos = appendData(pos, end, delta_monotonic);
-		pos = appendData(pos, end, "", 1);
-	}
-
 	if (req->upgraded()) {
 		pos = appendData(pos, end, P_STATIC_STRING_WITH_NULL("HTTP_CONNECTION"));
 		pos = appendData(pos, end, P_STATIC_STRING_WITH_NULL("upgrade"));
@@ -635,7 +617,7 @@ Controller::sendHeaderToAppWithHttpProtocol(Client *client, Request *req) {
 
 	cache.cached = false;
 
-	if (OXT_UNLIKELY(getLogLevel() >= LVL_DEBUG3)) {
+	if (OXT_UNLIKELY(LoggingKit::getLevel() >= LoggingKit::DEBUG3)) {
 		struct iovec *buffers;
 		unsigned int nbuffers, dataSize;
 		bool ok;
@@ -863,20 +845,6 @@ Controller::constructHeaderBuffersForHttpProtocol(Request *req, struct iovec *bu
 		PUSH_STATIC_BUFFER("\r\n");
 	}
 
-	if (req->options.analytics) {
-		PUSH_STATIC_BUFFER("!~Passenger-Txn-Id: ");
-
-		if (buffers != NULL) {
-			BEGIN_PUSH_NEXT_BUFFER();
-			buffers[i].iov_base = (void *) req->options.transaction->getTxnId().data();
-			buffers[i].iov_len  = req->options.transaction->getTxnId().size();
-		}
-		INC_BUFFER_ITER(i);
-		dataSize += req->options.transaction->getTxnId().size();
-
-		PUSH_STATIC_BUFFER("\r\n");
-	}
-
 	PUSH_STATIC_BUFFER("\r\n");
 
 	nbuffers = i;
