@@ -1,6 +1,6 @@
 /*
  *  Phusion Passenger - https://www.phusionpassenger.com/
- *  Copyright (c) 2010-2017 Phusion Holding B.V.
+ *  Copyright (c) 2010-2018 Phusion Holding B.V.
  *
  *  "Passenger", "Phusion Passenger" and "Union Station" are registered
  *  trademarks of Phusion Holding B.V.
@@ -38,6 +38,7 @@
 
 #include <boost/config.hpp>
 #include <boost/scoped_ptr.hpp>
+#include <boost/foreach.hpp>
 #include <sys/types.h>
 #include <sys/socket.h>
 #include <sys/stat.h>
@@ -86,10 +87,10 @@
 #include <ResourceLocator.h>
 #include <BackgroundEventLoop.cpp>
 #include <FileTools/FileManip.h>
+#include <FileTools/PathSecurityCheck.h>
 #include <Exceptions.h>
 #include <Utils.h>
 #include <Utils/Timer.h>
-#include <Utils/IOUtils.h>
 #include <Utils/MessageIO.h>
 #include <Utils/VariantMap.h>
 #include <Core/OptionParser.h>
@@ -98,7 +99,6 @@
 #include <Core/Config.h>
 #include <Core/ConfigChange.h>
 #include <Core/ApplicationPool/Pool.h>
-#include <Core/UnionStation/Context.h>
 #include <Core/SecurityUpdateChecker.h>
 #include <Core/AdminPanelConnector.h>
 
@@ -146,9 +146,9 @@ namespace Core {
 
 		ResourceLocator resourceLocator;
 		RandomGeneratorPtr randomGenerator;
-		UnionStation::ContextPtr unionStationContext;
-		SpawningKit::ConfigPtr spawningKitConfig;
-		SpawningKit::FactoryPtr spawningKitFactory;
+		SpawningKit::Context::Schema spawningKitContextSchema;
+		SpawningKit::ContextPtr spawningKitContext;
+		ApplicationPool2::ContextPtr appPoolContext;
 		PoolPtr appPool;
 		Json::Value singleAppModeConfig;
 
@@ -240,7 +240,7 @@ initializePrivilegedWorkingObjects() {
 	if (password.isString()) {
 		wo->controllerSecureHeadersPassword = password.asString();
 	} else if (password.isObject()) {
-		wo->controllerSecureHeadersPassword = strip(readAll(password["path"].asString()));
+		wo->controllerSecureHeadersPassword = strip(unsafeReadFile(password["path"].asString()));
 	}
 }
 
@@ -621,12 +621,6 @@ onTerminationSignal(EV_P_ struct ev_signal *watcher, int revents) {
 	}
 }
 
-static void
-spawningKitErrorHandler(const SpawningKit::ConfigPtr &config, SpawnException &e, const Options &options) {
-	ApplicationPool2::processAndLogNewSpawnException(e, options, config);
-}
-
-
 static void
 initializeCurl() {
 	TRACE_POINT();
@@ -658,30 +652,25 @@ initializeNonPrivilegedWorkingObjects() {
 	}
 
 	UPDATE_TRACE_POINT();
-	if (!coreConfig->get("ust_router_address").isNull()) {
-		wo->unionStationContext = boost::make_shared<UnionStation::Context>(
-			coreConfig->get("ust_router_address").asString(),
-			"logging",
-			coreConfig->get("ust_router_password").asString());
+	wo->spawningKitContext = boost::make_shared<SpawningKit::Context>(
+		wo->spawningKitContextSchema);
+	wo->spawningKitContext->resourceLocator = &wo->resourceLocator;
+	wo->spawningKitContext->randomGenerator = wo->randomGenerator;
+	wo->spawningKitContext->integrationMode = coreConfig->get("integration_mode").asString();
+	wo->spawningKitContext->instanceDir = coreConfig->get("instance_dir").asString();
+	if (!wo->spawningKitContext->instanceDir.empty()) {
+		wo->spawningKitContext->instanceDir = absolutizePath(
+			wo->spawningKitContext->instanceDir);
 	}
+	wo->spawningKitContext->finalize();
 
 	UPDATE_TRACE_POINT();
-	wo->spawningKitConfig = boost::make_shared<SpawningKit::Config>();
-	wo->spawningKitConfig->resourceLocator = &wo->resourceLocator;
-	wo->spawningKitConfig->agentConfig = coreConfig->inspectEffectiveValues();
-	wo->spawningKitConfig->errorHandler = spawningKitErrorHandler;
-	wo->spawningKitConfig->unionStationContext = wo->unionStationContext;
-	wo->spawningKitConfig->randomGenerator = wo->randomGenerator;
-	wo->spawningKitConfig->instanceDir = coreConfig->get("instance_dir").asString();
-	if (!wo->spawningKitConfig->instanceDir.empty()) {
-		wo->spawningKitConfig->instanceDir = absolutizePath(
-			wo->spawningKitConfig->instanceDir);
-	}
-	wo->spawningKitConfig->finalize();
-
-	UPDATE_TRACE_POINT();
-	wo->spawningKitFactory = boost::make_shared<SpawningKit::Factory>(wo->spawningKitConfig);
-	wo->appPool = boost::make_shared<Pool>(wo->spawningKitFactory, coreConfig->inspectEffectiveValues());
+	wo->appPoolContext = boost::make_shared<ApplicationPool2::Context>();
+	wo->appPoolContext->spawningKitFactory = boost::make_shared<SpawningKit::Factory>(
+		wo->spawningKitContext.get());
+	wo->appPoolContext->agentConfig = coreConfig->inspectEffectiveValues();
+	wo->appPoolContext->finalize();
+	wo->appPool = boost::make_shared<Pool>(wo->appPoolContext.get());
 	wo->appPool->initialize();
 	wo->appPool->setMax(coreConfig->get("max_pool_size").asInt());
 	wo->appPool->setMaxIdleTime(coreConfig->get("pool_idle_time").asInt() * 1000000ULL);
@@ -727,7 +716,6 @@ initializeNonPrivilegedWorkingObjects() {
 			coreSchema->controllerSingleAppMode.translator);
 		two.controller->resourceLocator = &wo->resourceLocator;
 		two.controller->appPool = wo->appPool;
-		two.controller->unionStationContext = wo->unionStationContext;
 		two.controller->shutdownFinishCallback = controllerShutdownFinished;
 		two.controller->initialize();
 		wo->shutdownCounter.fetch_add(1, boost::memory_order_relaxed);
@@ -863,6 +851,10 @@ initializeAdminPanelConnector() {
 	connector->resourceLocator = &wo.resourceLocator;
 	connector->appPool = wo.appPool;
 	connector->configGetter = inspectConfig;
+	for (unsigned int i = 0; i < wo.threadWorkingObjects.size(); i++) {
+		ThreadWorkingObjects *two = &wo.threadWorkingObjects[i];
+		connector->controllers.push_back(two->controller);
+	}
 	connector->initialize();
 	wo.shutdownCounter.fetch_add(1, boost::memory_order_relaxed);
 	wo.adminPanelConnector = connector;
@@ -894,48 +886,6 @@ prestartWebApps() {
 	);
 }
 
-/**
- * See warnIfPassengerRootVulnerable()
- */
-static void
-warnIfPathVulnerable(const char *path, string &warnings) {
-	struct stat pathStat;
-
-	if (stat(path, &pathStat) == -1) {
-		P_DEBUG("Vulnerability check skipped: stat error on " << path << " (errno: " << errno << ")");
-		return; // fatal: we need that stat for both checks below
-	}
-
-	// Non-root ownership
-	struct passwd pathOwner;
-	struct passwd *pwdResult;
-
-	boost::shared_array<char> strings;
-	long stringsBufSize = std::max<long>(1024 * 128, sysconf(_SC_GETPW_R_SIZE_MAX));
-	strings.reset(new char[stringsBufSize]);
-	errno = 0;
-	if (getpwuid_r(pathStat.st_uid, &pathOwner, strings.get(), stringsBufSize, &pwdResult) == -1) {
-		P_DEBUG("Vulnerability check (owner) skipped: getpwuid_r error on " << path << " (owner UID: " <<
-				pathStat.st_uid << ", errno: " << errno << ")");
-	} else if (pwdResult == NULL) {
-		P_DEBUG("Vulnerability check (owner) skipped: getpwuid_r empty on " << path << " (owner UID: " <<
-				pathStat.st_uid << ", errno: " << errno << ")");
-	} else if (pathOwner.pw_uid != 0) {
-		warnings.append("\nThe path \"");
-		warnings.append(path);
-		warnings.append("\" can be modified by user \"");
-		warnings.append(pathOwner.pw_name);
-		warnings.append("\" (or applications running as that user). Change the owner of the path to root, or avoid running Passenger as root.");
-	}
-
-	// World writeable access rights
-	if ((pathStat.st_mode & S_IWOTH) != 0) {
-		warnings.append("\nThe path \"");
-		warnings.append(path);
-		warnings.append("\" is writeable by any user (or application). Limit write access on the path to only the root user/group.");
-	}
-}
-
 /*
  * Emit a warning (log) if the Passenger root dir (and/or its parents) can be modified by non-root users
  * while Passenger was run as root (because non-root users can then tamper with something running as root).
@@ -955,18 +905,28 @@ warnIfPassengerRootVulnerable() {
 	}
 
 	string root = workingObjects->resourceLocator.getInstallSpec();
-	string checkPath = absolutizePath(root);
-	// Check the Passenger root and all dirs above it for ownership and world-writeability
-	string warnings;
-	while (!checkPath.empty() && checkPath != "/") {
-		warnIfPathVulnerable(checkPath.c_str(), warnings);
-
-		checkPath = extractDirName(checkPath);
-	}
-	if (!warnings.empty()) {
-		P_WARN("WARNING: potential privilege escalation vulnerability. "
-			PROGRAM_NAME " is running as root, and part(s) of the passenger root path ("
-			<< root << ") can be changed by non-root user(s):" << warnings);
+	vector<string> errors, checkErrors;
+	if (isPathProbablySecureForRootUse(root, errors, checkErrors)) {
+		if (!checkErrors.empty()) {
+			string message = "WARNING: unable to perform privilege escalation vulnerability detection:\n";
+			foreach (string line, checkErrors) {
+				message.append("\n - " + line);
+			}
+			P_WARN(message);
+		}
+	} else {
+		string message = "WARNING: potential privilege escalation vulnerability detected. " \
+			PROGRAM_NAME " is running as root, and part(s) of the " SHORT_PROGRAM_NAME
+			" root path (" + root + ") can be changed by non-root user(s):\n";
+		foreach (string line, errors) {
+			message.append("\n - " + line);
+		}
+		foreach (string line, checkErrors) {
+			message.append("\n - " + line);
+		}
+		message.append("\n\nPlease either fix up the permissions for the insecure paths, or install "
+			SHORT_PROGRAM_NAME " in a different location that can only be modified by root.");
+		P_WARN(message);
 	}
 }
 
@@ -1309,6 +1269,12 @@ parseOptions(int argc, const char *argv[], ConfigKit::Store &config) {
 	}
 }
 
+static void
+loggingKitPreInitFunc(Json::Value &loggingKitInitialConfig) {
+	loggingKitInitialConfig = manipulateLoggingKitConfig(*coreConfig,
+		loggingKitInitialConfig);
+}
+
 int
 coreMain(int argc, char *argv[]) {
 	int ret;
@@ -1317,7 +1283,7 @@ coreMain(int argc, char *argv[]) {
 	coreConfig = new ConfigKit::Store(*coreSchema);
 	initializeAgent(argc, &argv, SHORT_PROGRAM_NAME " core",
 		*coreConfig, coreSchema->loggingKit.translator,
-		parseOptions, NULL, 2);
+		parseOptions, loggingKitPreInitFunc, 2);
 
 #if !BOOST_OS_MACOS
 	restoreOomScore(coreConfig->get("oom_score").asString());
